package de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Utilities;

import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Exceptions.ParserException;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.PlayerStatistics;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Parser {

    /*
    Includes public static Parser methods for:
    -username + password fields when creating account or logging in with existing credentials
    -map name, when creating map
    -IP-address field when joining game
    -Map-name or map-ID, when searching for a map
    -Parser for mapData from newly created map (length, formatting (spaces), tile-types, border-tiles etc.)
    */

    public static void usernameValid(String username) throws ParserException {
        int minUsernameLength = 4;
        int maxUsernameLength = 40;

        if(username.length() < minUsernameLength){
            throw new ParserException("Username too short! Min length is " + minUsernameLength + " characters.");
        }
        if(username.length() > maxUsernameLength){
            throw new ParserException("Username too long! Max length is " + maxUsernameLength + " characters.");
        }

        Pattern pat = Pattern.compile("[^a-z0-9_-]", Pattern.CASE_INSENSITIVE);
        Matcher mat = pat.matcher(username);
        boolean result = mat.find();
        if(result){
            throw new ParserException("Forbidden characters used! Only user letters A-Z or numbers 0-9 or symbols '-' or '_' !");
        }
    }

    public static void passwordValid(String password) throws ParserException {
        int minPasswordLength = 4;
        int maxPasswordLength = 40;

        if(password.length() < minPasswordLength){
            throw new ParserException("Password too short! Min length is " + minPasswordLength + " characters.");
        }
        if(password.length() > maxPasswordLength){
            throw new ParserException("Password too long! Max length is " + maxPasswordLength + " characters.");
        }

        Pattern pat = Pattern.compile("[^a-z0-9_-]", Pattern.CASE_INSENSITIVE);
        Matcher mat = pat.matcher(password);
        boolean result = mat.find();
        if(result){
            throw new ParserException("Forbidden characters used! Only user letters A-Z or numbers 0-9 or symbols '-' or '_' !");
        }
    }

    public static void mapNameValid(String mapName) throws ParserException {
        int minMapNameLength = 3;
        int maxMapNameLength = 30;

        if(mapName.length() < minMapNameLength){
            throw new ParserException("Map-Name too short! Min length is " + minMapNameLength + " characters.");
        }
        if(mapName.length() > maxMapNameLength){
            throw new ParserException("Map-Name too long! Max length is " + maxMapNameLength + " characters.");
        }

        Pattern pat = Pattern.compile("[^a-z0-9_-]", Pattern.CASE_INSENSITIVE);
        Matcher mat = pat.matcher(mapName);
        boolean result = mat.find();
        if(result){
            throw new ParserException("Forbidden characters used! Only user letters A-Z or numbers 0-9 or symbols '-' or '_' !");
        }
    }

    public static void sha1HexHashValid(String mapID) throws ParserException {
        int hashLength = 40;

        if(mapID.length() != hashLength){
            throw new ParserException("SHA1 hash length not correct. Must have length of " + hashLength + " characters!");
        }

        Pattern pat = Pattern.compile("[^a-f0-9]");         //maybe make case-insensitive -> letters can be uppercase and still be valid hex number
        Matcher mat = pat.matcher(mapID);
        boolean result = mat.find();
        if(result){
            throw new ParserException("Forbidden characters used! SHA1 hash only consists out of letters a-f and/or numbers 0-9!");
        }
    }

    public static void ipAddressValid(String address) throws ParserException {

        int addressMinLength = 7;
        int addressMaxLength = 15;
        int addressLength = address.length();
        int octetMinDecValue = 0;
        int octetMaxDecValue = 255;

        if(address.equals("localhost")){
            return;
        }

        if(addressLength > addressMaxLength){
            throw new ParserException("IP-Address too long. Must have maximum length of " + addressMaxLength + " characters including octet dividers (.) !");
        }

        if(addressLength < addressMinLength){
            throw new ParserException("IP-Address too short. Must have minimum length of " + addressMinLength + " characters including octet dividers (.) !");
        }

        if(address.charAt(0) == '.' | address.charAt(addressLength - 1) == '.'){
            throw new ParserException("IP-Address must not start nor end with octet dividers (.) !");
        }

        int counter = 0;
        for(int i = 0; addressLength > i ; i++){
            if(address.charAt(i) == '.'){
                counter++;
            }
        }
        if(counter != 3){
            throw new ParserException("IP-Address must contain exactly 3 octet dividers (.) !");
        }

        for(int i = 0; addressLength > i ; i++){
            if(address.charAt(i) == '.' && address.charAt(i + 1) == '.'){
                throw new ParserException("IP-Address must not contain two or more consecutive dividers (.) without number in between!");
            }
        }

        Pattern pat = Pattern.compile("[^0-9.]");
        for(int i = 0; addressLength > i ; i++){
            Matcher mat = pat.matcher(address.substring(i, i+1));
            boolean result = mat.find();
            if(result){
                throw new ParserException("IP address must consist out of numbers 0-9 and separators (.) !");
            }
        }

        counter = 0;
        int[] positions = new int[4];
        positions[0] = 0;
        for(int i = 0; addressLength > i ; i++){
            if(address.charAt(i) == '.'){
                counter++;
                positions[counter] = i + 1;
            }
        }
        for(int i = 0; positions.length > i; i++){
            if(positions.length - 1 > i) {
                if (Integer.parseInt(address.substring(positions[i], positions[i + 1] - 1)) > octetMaxDecValue) {
                    throw new ParserException("Octets of IP-address must not exceed 255!");
                }
            }
            else{
                if (Integer.parseInt(address.substring(positions[i], addressLength)) > octetMaxDecValue) {
                    throw new ParserException("Octets of IP-address must not exceed 255!");
                }
            }
        }
    }

    public static void mapDataValid(String mapData) throws ParserException {

        //values could be handed to method (Instance of MapData Class)
        int actualMapDataLength = mapData.length();
        int mapWidth = 18;         //Tiles
        int mapHeight = 18;        //Tiles
        int spacesAmount = (mapHeight * mapWidth) - 1;      //space used as separator
        int totalTilesAmount = mapHeight * mapWidth;
        int borderTilesAmount = (2 * mapWidth) + (2 * (mapHeight - 2));
        int spawnTilesAmount = 4;
        int fieldTilesAmount = totalTilesAmount - borderTilesAmount - spawnTilesAmount;
        int maxTileNumberDigits = 2;  //Tile-Numbers are max 2 digits: 10 to 29
        int minTileNumberDigits = 1; //Tile-Numbers are min 1 digit: 0 to 9
        int minTileNumber = 0;
        int maxTileNumber = 29;
        int walkableTileMinNumber = 0;
        int walkableTileMaxNumber = 9;
        int indestructibleTileMinNumber = 10;
        int indestructibleTileMaxNumber = 19;
        int destructibleTileMinNumber = 20;
        int destructibleTileMaxNumber = 29;
        int maxMapDataLength = spacesAmount + borderTilesAmount * Integer.toString(indestructibleTileMaxNumber).length() + spawnTilesAmount * Integer.toString(walkableTileMaxNumber).length() + fieldTilesAmount * Integer.toString(destructibleTileMaxNumber).length();           //"Integer.toString(destructibleTileMaxNumber).length()" results in number of digits of destructibleTileMaxNumber
        int minMapDataLength = spacesAmount + borderTilesAmount * Integer.toString(indestructibleTileMinNumber).length() + spawnTilesAmount * Integer.toString(walkableTileMinNumber).length() + fieldTilesAmount * Integer.toString(walkableTileMinNumber).length();

        if(actualMapDataLength > maxMapDataLength){
            throw new ParserException("Map-Data corrupted - mapData String must not exceed length of " + maxMapDataLength + " chars including spaces for " + mapWidth + "x" + mapHeight + " Tiles map!");
        }

        if(actualMapDataLength < minMapDataLength){
            throw new ParserException("Map-Data corrupted - minimum length of mapData String must be " + minMapDataLength + " chars including spaces for " + mapWidth + "x" + mapHeight + " Tiles map!");
        }

        if(mapData.charAt(0) == ' ' | mapData.charAt(actualMapDataLength - 1) == ' '){
            throw new ParserException("Map-Data corrupted - must not start nor end with space!");
        }

        for(int i = 0; actualMapDataLength > i ; i++){
            if(mapData.charAt(i) == ' ' && mapData.charAt(i + 1) == ' '){
                throw new ParserException("Map-Data corrupted - must not contain two or more consecutive spaces without number in between!");
            }
        }

        int counter = 0;
        for(int i = 0; i < actualMapDataLength; i++){
            if((mapData.charAt(i) == ' ')){
                counter++;
            }
        }
        if(counter != spacesAmount){
            throw new ParserException("Map-Data corrupted - number of spaces must be " + spacesAmount + "!");
        }

        Pattern pat = Pattern.compile("[^0-9 ]");
        for(int i = 0; actualMapDataLength > i ; i++){
            Matcher mat = pat.matcher(mapData.substring(i, i+1));
            boolean result = mat.find();
            if(result){
                throw new ParserException("Map-Data corrupted - must consist out of numbers 0-9 and space as separator!");
            }
        }

        counter = 0;
        int[] positions = new int[totalTilesAmount];
        positions[0] = 0;
        for(int i = 0; actualMapDataLength > i ; i++){
            if(mapData.charAt(i) == ' '){
                counter++;
                positions[counter] = i + 1;
            }
        }
        for(int i = 0; positions.length > i; i++){
            if(positions.length - 1 > i) {
                if (Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) > maxTileNumber | Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) < minTileNumber) {
                    throw new ParserException("Map-Data corrupted - Tile number must be between " + minTileNumber + " and " + maxTileNumber + "!");
                }
            }
            else{
                if (Integer.parseInt(mapData.substring(positions[i], actualMapDataLength)) > maxTileNumber | Integer.parseInt(mapData.substring(positions[i], actualMapDataLength)) < minTileNumber) {
                    throw new ParserException("Map-Data corrupted - Tile number must be between " + minTileNumber + " and " + maxTileNumber + "!");
                }
            }
        }

        //Alternative code for checking if tile number is between minTileNumber and maxTileNumber:

        /*
        String combinedNumbers = "";
        for(int i = 0; i < actualMapDataLength - 1; i++){
            if(mapData.charAt(i+1) == ' '){
                combinedNumbers = combinedNumbers + mapData.charAt(i);
                if((Integer.parseInt(combinedNumbers) > maxTileNumber) | (Integer.parseInt(combinedNumbers) < minTileNumber)){
                    throw new ParserException("Map-Data corrupted - Tile number must be between " + minTileNumber + " and " + maxTileNumber + "!");
                }
                combinedNumbers = "";
                if(i < actualMapDataLength - 1){
                    i++;
                }
            }
            else if(i >= actualMapDataLength - 2){
                combinedNumbers = combinedNumbers + mapData.charAt(i) + mapData.charAt(i+1);
                if((Integer.parseInt(combinedNumbers) > maxTileNumber) | (Integer.parseInt(combinedNumbers) < minTileNumber)){
                    throw new ParserException("Map-Data corrupted - Tile number must be between " + minTileNumber + " and " + maxTileNumber + "!");
                }
            }
            else{
                combinedNumbers = combinedNumbers + mapData.charAt(i);
            }
        }
        */

        //corner Tiles included
        for(int i = 0; i < mapWidth; i++){
            if(Integer.parseInt(mapData.substring(positions[i], positions[i+1] - 1)) > indestructibleTileMaxNumber | Integer.parseInt(mapData.substring(positions[i], positions[i+1] - 1)) < indestructibleTileMinNumber){
                throw new ParserException("Map-Data corrupted - Top-Line must be border tiles!");
            }
        }

        //corner Tiles included
        for(int i = positions.length - mapWidth; i < positions.length; i++){
            if(positions.length - 1 > i) {
                if (Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) > indestructibleTileMaxNumber | Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) < indestructibleTileMinNumber) {
                    throw new ParserException("Map-Data corrupted - Bottom-Line must be border tiles!");
                }
            }
            else{
                if (Integer.parseInt(mapData.substring(positions[i], actualMapDataLength)) > indestructibleTileMaxNumber | Integer.parseInt(mapData.substring(positions[i], actualMapDataLength)) < indestructibleTileMinNumber) {
                    throw new ParserException("Map-Data corrupted - Bottom-Line must be border tiles!");
                }
            }
        }

        //corner Tiles excluded
        for(int i = mapWidth; i < positions.length - mapWidth; i = i + mapWidth){
            if(i % mapWidth == 0 && Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) > indestructibleTileMaxNumber | Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) < indestructibleTileMinNumber){
                throw new ParserException("Map-Data corrupted - Left Edge must be border tiles!");
            }
        }

        //corner Tiles excluded
        for(int i = mapWidth - 1; i < positions.length - mapWidth; i = i + mapWidth){
            if((i + 1) % mapWidth == 0 && Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) > indestructibleTileMaxNumber | Integer.parseInt(mapData.substring(positions[i], positions[i + 1] - 1)) < indestructibleTileMinNumber){
                throw new ParserException("Map-Data corrupted - Right Edge must be border tiles!");
            }
        }

        if(Integer.parseInt(mapData.substring(positions[mapWidth + 1], positions[mapWidth + 2] - 1)) < walkableTileMinNumber | Integer.parseInt(mapData.substring(positions[mapWidth + 1], positions[mapWidth + 2] - 1)) > walkableTileMaxNumber){
            throw new ParserException("Map-Data corrupted - Player spawn top left must use walkable tile!");
        }

        if(Integer.parseInt(mapData.substring(positions[(mapWidth*2) - 2], positions[(mapWidth*2) - 1] - 1)) < walkableTileMinNumber | Integer.parseInt(mapData.substring(positions[(mapWidth*2) - 2], positions[(mapWidth*2) - 1] - 1)) > walkableTileMaxNumber){
            throw new ParserException("Map-Data corrupted - Player spawn top right must use walkable tile!");
        }

        if(Integer.parseInt(mapData.substring(positions[positions.length - (mapWidth*2) + 1], positions[positions.length - (mapWidth*2) + 2] - 1)) < walkableTileMinNumber | Integer.parseInt(mapData.substring(positions[positions.length - (mapWidth*2) + 1], positions[positions.length - (mapWidth*2) + 2] - 1)) > walkableTileMaxNumber) {
            throw new ParserException("Map-Data corrupted - Player spawn bottom left must use walkable tile!");
        }

        if(Integer.parseInt(mapData.substring(positions[positions.length - mapWidth - 2], positions[positions.length - mapWidth - 1] - 1)) < walkableTileMinNumber | Integer.parseInt(mapData.substring(positions[positions.length - mapWidth - 2], positions[positions.length - mapWidth - 1] - 1)) > walkableTileMaxNumber){
            throw new ParserException("Map-Data corrupted - Player spawn bottom right must use walkable tile!");
        }
    }

    public static void playerStatsValid(PlayerStatistics stats) throws ParserException{

        if(stats.getGamesLost() < 0){
            throw new ParserException("Player statistics data corrupted - Minimum value for lost games is 0!");
        }

        if(stats.getGamesWon() < 0){
            throw new ParserException("Player statistics data corrupted - Minimum value for won games is 0!");
        }

        if(stats.getKills() < 0){
            throw new ParserException("Player statistics data corrupted - Minimum value for kills is 0!");
        }

        if(stats.getDeaths() < 0){
            throw new ParserException("Player statistics data corrupted - Minimum value for deaths is 0!");
        }

        if(stats.getBlocksDestroyed() < 0){
            throw new ParserException("Player statistics data corrupted - Minimum value for destroyed blocks is 0!");
        }

        if(stats.getGameTime() < 0){
            throw new ParserException("Player statistics data corrupted - Minimum value for ingame time is 0 seconds!");
        }
    }

    public static void volumeValid(int volume) throws ParserException{

        if(volume > 100 | volume < 0){
            throw new ParserException("Volume must be between 0 and 100!");
        }

    }

}
